@@ -243,7 +243,7 @@ static int dump_version_two(FILE *file, FILE *out)
 		     "%llu\t"
 		     "%u\t"
 		     "%u\t"
-		     "%u\t",
+		     "%u",
 		     &num_prmode,
 		     &num_exmode,
 		     &num_prmode_failed,
@@ -275,7 +275,8 @@ out:
 
 /* 0 = eof, > 0 = success, < 0 = error */
 static int dump_version_one(FILE *file, FILE *out, int lvbs, int only_busy,
-			    int *skipped)
+			    struct list_head *locklist, int *skipped,
+			    unsigned int version)
 {
 	char id[OCFS2_LOCK_ID_MAX_LEN + 1];	
 	char lvb[DLM_LVB_LEN];
@@ -315,7 +316,7 @@ static int dump_version_one(FILE *file, FILE *out, int lvbs, int only_busy,
 		 * include the field delimiting '\t' then fscanf will
 		 * also catch the record delimiting '\n' character,
 		 * which we want to save for the caller to find. */
-		if (i == (DLM_LVB_LEN - 1))
+		if ((i == (DLM_LVB_LEN - 1)) && (version < 2))
 			format = "0x%x";
 
 		ret = fscanf(file, format, &dummy);
@@ -327,6 +328,13 @@ static int dump_version_one(FILE *file, FILE *out, int lvbs, int only_busy,
 		lvb[i] = (char) dummy;
 	}
 
+	if (!list_empty(locklist)) {
+		if (!del_from_stringlist(id, locklist)) {
+			ret = 1;
+			goto out;
+		}
+	}
+
 	if (only_busy) {
 		if (!(flags & OCFS2_LOCK_BUSY)) {
 			ret = 1;
@@ -370,7 +378,8 @@ static int end_line(FILE *f)
 
 #define CURRENT_PROTO 2
 /* returns 0 on error or end of file */
-static int dump_one_lockres(FILE *file, FILE *out, int lvbs, int only_busy)
+static int dump_one_lockres(FILE *file, FILE *out, int lvbs, int only_busy,
+			    struct list_head *locklist)
 {
 	unsigned int version;
 	int ret;
@@ -386,20 +395,19 @@ static int dump_one_lockres(FILE *file, FILE *out, int lvbs, int only_busy)
 		return 0;
 	}
 
-	ret = dump_version_one(file, out, lvbs, only_busy, &skipped);
+	ret = dump_version_one(file, out, lvbs, only_busy, locklist, &skipped, version);
 	if (ret <= 0)
 		return 0;
 
 	if (!skipped) {
-		if (version >= 2) {
+		if (version == CURRENT_PROTO) {
 			ret = dump_version_two(file, out);
 			if (ret <= 0)
 				return 0;
 		}
-	}
 
-	if (!skipped)
 		fprintf(out, "\n");
+	}
 
 	/* Read to the end of the record here. Any new fields tagged
 	 * onto the current format will be silently ignored. */
@@ -408,30 +416,45 @@ static int dump_one_lockres(FILE *file, FILE *out, int lvbs, int only_busy)
 	return ret;
 }
 
-void dump_fs_locks(char *uuid_str, FILE *out, int dump_lvbs, int only_busy)
+void dump_fs_locks(char *uuid_str, FILE *out, char *path, int dump_lvbs,
+		   int only_busy, struct list_head *locklist)
 {
 	errcode_t ret;
 	char debugfs_path[PATH_MAX];
 	FILE *file;
+	int show_select;
 
-	ret = get_debugfs_path(debugfs_path, sizeof(debugfs_path));
-	if (ret) {
-		fprintf(stderr, "Could not locate debugfs file system. "
-			"Perhaps it is not mounted?\n");
-		return;
+	if (path == NULL) {
+		ret = get_debugfs_path(debugfs_path, sizeof(debugfs_path));
+		if (ret) {
+			fprintf(stderr, "Could not locate debugfs file system. "
+				"Perhaps it is not mounted?\n");
+			return;
+		}
+
+		ret = open_debugfs_file(debugfs_path, "ocfs2", uuid_str,
+					"locking_state", &file);
+		if (ret) {
+			fprintf(stderr, "Could not open debug state for "
+				"\"%s\".\nPerhaps that OCFS2 file system is "
+				"not mounted?\n", uuid_str);
+			return;
+		}
+	} else {
+		file = fopen(path, "r");
+		if (!file) {
+			fprintf(stderr, "Could not open file at \"%s\"\n",
+				path);
+			return;
+		}
 	}
 
-	ret = open_debugfs_file(debugfs_path, "ocfs2", uuid_str,
-				"locking_state", &file);
-	if (ret) {
-		fprintf(stderr, "Could not open debug state for \"%s\".\n"
-			"Perhaps that OCFS2 file system is not mounted?\n",
-			uuid_str);
-		return;
-	}
+	show_select = !list_empty(locklist);
 
-	while (dump_one_lockres(file, out, dump_lvbs, only_busy))
-		;
+	while (dump_one_lockres(file, out, dump_lvbs, only_busy, locklist)) {
+		if (show_select && list_empty(locklist))
+			break;
+	}
 
 	fclose(file);
 }
