@@ -1,4 +1,6 @@
-/*
+/* -*- mode: c; c-basic-offset: 8; -*-
+ * vim: noexpandtab sw=8 ts=8 sts=0:
+ *
  * journal.c
  *
  * reads the journal file
@@ -27,12 +29,14 @@
 
 extern dbgfs_gbls gbls;
 
-static void scan_journal(FILE *out, char *buf, int len,
-			 uint64_t *blocknum, uint64_t *last_unknown)
+static enum dump_block_type detect_block (char *buf);
+
+static void scan_journal(FILE *out, journal_superblock_t *jsb, char *buf,
+			 int len, uint64_t *blocknum, uint64_t *last_unknown)
 {
 	char *block;
 	char *p;
-	int type;
+	enum dump_block_type type;
 	journal_header_t *header;
 
 	p = buf;
@@ -40,15 +44,15 @@ static void scan_journal(FILE *out, char *buf, int len,
 	while (len) {
 		block = p;
 		header = (journal_header_t *)block;
-		if (header->h_magic == ntohl(JFS_MAGIC_NUMBER)) {
+		if (header->h_magic == ntohl(JBD2_MAGIC_NUMBER)) {
 			if (*last_unknown) {
 				dump_jbd_unknown(out, *last_unknown, *blocknum);
 				*last_unknown = 0;
 			}
-			dump_jbd_block(out, header, *blocknum);
+			dump_jbd_block(out, jsb, header, *blocknum);
 		} else {
 			type = detect_block(block);
-			if (type < 0) {
+			if (type == DUMP_BLOCK_UNKNOWN) {
 				if (*last_unknown == 0)
 					*last_unknown = *blocknum;
 			} else {
@@ -71,6 +75,7 @@ static void scan_journal(FILE *out, char *buf, int len,
 errcode_t read_journal(ocfs2_filesys *fs, uint64_t blkno, FILE *out)
 {
 	char *buf = NULL;
+	char *jsb_buf = NULL;
 	char *p;
 	uint64_t blocknum;
 	uint64_t len;
@@ -81,6 +86,7 @@ errcode_t read_journal(ocfs2_filesys *fs, uint64_t blkno, FILE *out)
 	int buflenbits;
 	ocfs2_cached_inode *ci = NULL;
 	errcode_t ret;
+	journal_superblock_t *jsb;
 
 	ret = ocfs2_read_cached_inode(fs, blkno, &ci);
 	if (ret) {
@@ -88,6 +94,13 @@ errcode_t read_journal(ocfs2_filesys *fs, uint64_t blkno, FILE *out)
 		goto bail;
 	}
 
+	ret = ocfs2_malloc_block(fs->fs_io, &jsb_buf);
+	if (ret) {
+		com_err(gbls.cmd, ret,
+			"while allocating journal superblock buffer");
+		goto bail;
+	}
+
 	buflenbits = buflen >>
 			OCFS2_RAW_SB(gbls.fs->fs_super)->s_blocksize_bits;
 	ret = ocfs2_malloc_blocks(fs->fs_io, buflenbits, &buf);
@@ -98,6 +111,7 @@ errcode_t read_journal(ocfs2_filesys *fs, uint64_t blkno, FILE *out)
 
 	offset = 0;
 	blocknum = 0;
+	jsb = (journal_superblock_t *)jsb_buf;
 	while (1) {
 		ret = ocfs2_file_read(ci, buf, buflen, offset, &got);
 		if (ret) {
@@ -112,13 +126,15 @@ errcode_t read_journal(ocfs2_filesys *fs, uint64_t blkno, FILE *out)
 		len = got;
 
 		if (offset == 0) {
-			dump_jbd_superblock(out, (journal_superblock_t *)buf);
+			memcpy(jsb_buf, buf, fs->fs_blocksize);
+			dump_jbd_superblock(out, jsb);
+			ocfs2_swap_journal_superblock(jsb);
 			blocknum++;
 			p += fs->fs_blocksize;
 			len -= fs->fs_blocksize;
 		}
 
-		scan_journal(out, p, len, &blocknum, &last_unknown);
+		scan_journal(out, jsb, p, len, &blocknum, &last_unknown);
 
 		if (got < buflen)
 			break;
@@ -131,6 +147,8 @@ errcode_t read_journal(ocfs2_filesys *fs, uint64_t blkno, FILE *out)
 	}
 
 bail:
+	if (jsb_buf)
+		ocfs2_free(&jsb_buf);
 	if (buf)
 		ocfs2_free(&buf);
 	if (ci)
@@ -143,31 +161,39 @@ bail:
  * detect_block()
  *
  */
-int detect_block (char *buf)
+static enum dump_block_type detect_block (char *buf)
 {
 	struct ocfs2_dinode *inode;
 	struct ocfs2_extent_block *extent;
 	struct ocfs2_group_desc *group;
-	int ret = -1;
+	struct ocfs2_dir_block_trailer *trailer;
+	enum dump_block_type ret = DUMP_BLOCK_UNKNOWN;
 
 	inode = (struct ocfs2_dinode *)buf;
-	if (!memcmp(inode->i_signature, OCFS2_INODE_SIGNATURE,
-		    sizeof(OCFS2_INODE_SIGNATURE))) {
-		ret = 1;
+	if (!strncmp((char *)inode->i_signature, OCFS2_INODE_SIGNATURE,
+		     sizeof(inode->i_signature))) {
+		ret = DUMP_BLOCK_INODE;
 		goto bail;
 	}
 
 	extent = (struct ocfs2_extent_block *)buf;
-	if (!memcmp(extent->h_signature, OCFS2_EXTENT_BLOCK_SIGNATURE,
-		    sizeof(OCFS2_EXTENT_BLOCK_SIGNATURE))) {
-		ret = 2;
+	if (!strncmp((char *)extent->h_signature, OCFS2_EXTENT_BLOCK_SIGNATURE,
+		     sizeof(extent->h_signature))) {
+		ret = DUMP_BLOCK_EXTENT_BLOCK;
 		goto bail;
 	}
 
 	group = (struct ocfs2_group_desc *)buf;
-	if (!memcmp(group->bg_signature, OCFS2_GROUP_DESC_SIGNATURE,
-		    sizeof(OCFS2_GROUP_DESC_SIGNATURE))) {
-		ret = 3;
+	if (!strncmp((char *)group->bg_signature, OCFS2_GROUP_DESC_SIGNATURE,
+		     sizeof(group->bg_signature))) {
+		ret = DUMP_BLOCK_GROUP_DESCRIPTOR;
+		goto bail;
+	}
+
+	trailer = ocfs2_dir_trailer_from_block(gbls.fs, buf);
+	if (!strncmp((char *)trailer->db_signature, OCFS2_DIR_TRAILER_SIGNATURE,
+		     sizeof(trailer->db_signature))) {
+		ret = DUMP_BLOCK_DIR_BLOCK;
 		goto bail;
 	}
 
