@@ -22,21 +22,45 @@
 
 #include "fsck.h"
 
+typedef errcode_t (check_leaf_er_func)(o2fsck_state *ost,
+				       uint64_t owner,
+				       struct ocfs2_extent_list *el,
+				       struct ocfs2_extent_rec *er,
+				       int *changed,
+				       void *para);
+typedef errcode_t (mark_leaf_er_alloc_func)(o2fsck_state *ost,
+					    struct ocfs2_extent_rec *er,
+					    uint32_t clusters,
+					    void *para);
+
 struct extent_info {
-	uint64_t	ei_max_size;
-	uint64_t	ei_clusters;
-	uint64_t	ei_last_eb_blk;
-	uint16_t	ei_expected_depth;
-	unsigned	ei_expect_depth:1;
+	uint64_t		ei_max_size;
+	uint64_t		ei_clusters;
+	uint64_t		ei_last_eb_blk;
+	uint16_t		ei_expected_depth;
+	unsigned		ei_expect_depth:1;
+	check_leaf_er_func	*chk_rec_func;
+	mark_leaf_er_alloc_func	*mark_rec_alloc_func;
+	void			*para;
 };
 
 errcode_t o2fsck_check_extents(o2fsck_state *ost,
                                struct ocfs2_dinode *di);
 
 errcode_t check_el(o2fsck_state *ost, struct extent_info *ei,
-			  struct ocfs2_dinode *di,
-			  struct ocfs2_extent_list *el,
-			  uint16_t max_recs, int *changed);
+		   uint64_t owner,
+		   struct ocfs2_extent_list *el,
+		   uint16_t max_recs, int *changed);
+errcode_t o2fsck_check_extent_rec(o2fsck_state *ost,
+				  uint64_t owner,
+				  struct ocfs2_extent_list *el,
+				  struct ocfs2_extent_rec *er,
+				  int *changed,
+				  void *para);
 
+errcode_t o2fsck_mark_tree_clusters_allocated(o2fsck_state *ost,
+					      struct ocfs2_extent_rec *rec,
+					      uint32_t clusters,
+					      void *para);
 #endif /* __O2FSCK_EXTENT_H__ */
 
