@@ -37,6 +37,7 @@ enum fsck_type
 	EB_GEN,
 	EB_GEN_FIX,
 	EXTENT_MARKED_UNWRITTEN,
+	EXTENT_MARKED_REFCOUNTED,
 	EXTENT_BLKNO_UNALIGNED,
 	EXTENT_CLUSTERS_OVERRUN,
 	EXTENT_EB_INVALID,
@@ -65,6 +66,16 @@ enum fsck_type
 	CHAIN_LINK_MAGIC,
 	CHAIN_LINK_RANGE,
 	CHAIN_BITS,
+	DISCONTIG_BG_DEPTH,
+	DISCONTIG_BG_COUNT,
+	DISCONTIG_BG_REC_RANGE,
+	DISCONTIG_BG_CORRUPT_LEAVES,
+	DISCONTIG_BG_CLUSTERS,
+	DISCONTIG_BG_LESS_CLUSTERS,
+	DISCONTIG_BG_NEXT_FREE_REC,
+	DISCONTIG_BG_LIST_CORRUPT,
+	DISCONTIG_BG_REC_CORRUPT,
+	DISCONTIG_BG_LEAF_CLUSTERS,
 	INODE_ALLOC_REPAIR,
 	INODE_SUBALLOC,
 	LALLOC_SIZE,
@@ -83,7 +94,6 @@ enum fsck_type
 	INODE_GEN,
 	INODE_GEN_FIX,
 	INODE_BLKNO,
-	INODE_LINK_NOT_CONNECTED,
 	ROOT_NOTDIR,
 	INODE_NZ_DTIME,
 	LINK_FAST_DATA,
@@ -129,6 +139,29 @@ enum fsck_type
 	JOURNAL_UNKNOWN_FEATURE,
 	JOURNAL_MISSING_FEATURE,
 	JOURNAL_TOO_SMALL,
+	QMAGIC_INVALID,
+	QTREE_BLK_INVALID,
+	DQBLK_INVALID,
+	DUP_DQBLK_INVALID,
+	DUP_DQBLK_VALID,
+	REFCOUNT_FLAG_INVALID,
+	REFCOUNT_LOC_INVALID,
+	RB_BLKNO,
+	RB_GEN,
+	RB_GEN_FIX,
+	RB_PARENT,
+	REFCOUNT_BLOCK_INVALID,
+	REFCOUNT_ROOT_BLOCK_INVALID,
+	REFCOUNT_LIST_COUNT,
+	REFCOUNT_LIST_USED,
+	REFCOUNT_CLUSTER_RANGE,
+	REFCOUNT_CLUSTER_COLLISION,
+	REFCOUNT_LIST_EMPTY,
+	REFCOUNT_CLUSTERS,
+	REFCOUNT_COUNT,
+	REFCOUNT_REC_REDUNDANT,
+	REFCOUNT_COUNT_INVALID,
+	DUP_CLUSTERS_ADD_REFCOUNT,
 	NUM_FSCK_TYPE
 };
 
@@ -162,9 +195,9 @@ enum fsck_type
  *
  * Inode field error: 	INODE_SUBALLOC, INODE_GEN, INODE_GEN_FIX,INODE_BLKNO,
 			INODE_NZ_DTIME, INODE_SIZE, INODE_SPARSE_SIZE,
-*			INODE_CLUSTERS, INODE_SPARSE_CLUSTERS, INODE_COUNT
+ *			INODE_CLUSTERS, INODE_SPARSE_CLUSTERS, INODE_COUNT
  *
- * Inode link not connected error: INODE_LINK_NOT_CONNECTED 
+ * Inode link not connected error:	INODE_NOT_CONNECTED
  *
  * Inode orphaned error:	INODE_ORPHANED
  *
@@ -186,9 +219,13 @@ enum fsck_type
  *
  * Special files error: ROOT_NOTDIR, ROOT_DIR_MISSING, LOSTFOUND_MISSING,
  *			DIR_DOTDOT
+ *
  * Journal file error:	JOURNAL_FILE_INVALID, JOURNAL_UNKNOWN_FEATURE,
-			JOURNAL_MISSING_FEATURE, JOURNAL_TOO_SMALL.
- * 	
+ *			JOURNAL_MISSING_FEATURE, JOURNAL_TOO_SMALL.
+ *
+ * Quota file error:	QMAGIC_INVALID, QTREE_BLK_INVALID, DQBLK_INVALID
+ *			DUP_DQBLK_INVALID, DUP_DQBLK_VALID
+ *
  * Link file error: LINK_FAST_DATA, LINK_NULLTERM, LINK_SIZE, LINK_BLOCKS
  *
  * Directory inode error: DIR_ZERO
@@ -210,5 +247,14 @@ enum fsck_type
  * Duplicate clusters error:	DUP_CLUSTERS_CLONE, DUP_CLUSTERS_DELETE
  *				DUP_CLUSTERS_SYSFILE_CLONE
  *
+ * refcount tree error:	EXTENT_MARKED_REFCOUNTED, REFCOUNT_FLAG_INVALID,
+ *			REFCOUNT_LOC_INVALID, RB_BLKNO,	RB_GEN,	RB_GEN_FIX,
+ *			RB_PARENT, REFCOUNT_BLOCK_INVALID,
+ *			REFCOUNT_ROOT_BLOCK_INVALID, REFCOUNT_LIST_COUNT,
+ *			REFCOUNT_LIST_USED, REFCOUNT_CLUSTER_RANGE,
+ *			REFCOUNT_CLUSTER_COLLISION, REFCOUNT_LIST_EMPTY
+ *			REFCOUNT_CLUSTERS, REFCOUNT_COUNT,
+ *			REFCOUNT_REC_REDUNDANT, REFCOUNT_COUNT_INVALID
+ *			DUP_CLUSTERS_ADD_REFCOUNT
  */
 #endif
