@@ -48,6 +48,18 @@ struct tunefs_flag_name {
 	uint16_t	tfn_flag;
 };
 
+/* Printable names for extent flags */
+struct extent_flag_name {
+	const char	*efn_name;
+	uint8_t		efn_flag;
+};
+
+/* Printable names for refcount flags */
+struct refcount_flag_name {
+	const char	*rfn_name;
+	uint32_t	rfn_flag;
+};
+
 struct feature_level_translation {
 	const char *fl_str;
 	enum ocfs2_feature_levels fl_type;
@@ -63,7 +75,8 @@ static struct feature_level_translation ocfs2_feature_levels_table[] = {
 static ocfs2_fs_options feature_level_defaults[] = {
 	{OCFS2_FEATURE_COMPAT_BACKUP_SB | OCFS2_FEATURE_COMPAT_JBD2_SB,
 	 OCFS2_FEATURE_INCOMPAT_SPARSE_ALLOC |
-	 OCFS2_FEATURE_INCOMPAT_INLINE_DATA,
+	 OCFS2_FEATURE_INCOMPAT_INLINE_DATA |
+	 OCFS2_FEATURE_INCOMPAT_XATTR,
 	 OCFS2_FEATURE_RO_COMPAT_UNWRITTEN},  /* OCFS2_FEATURE_LEVEL_DEFAULT */
 
 	{OCFS2_FEATURE_COMPAT_BACKUP_SB | OCFS2_FEATURE_COMPAT_JBD2_SB,
@@ -75,8 +88,40 @@ static ocfs2_fs_options feature_level_defaults[] = {
 	 OCFS2_FEATURE_INCOMPAT_EXTENDED_SLOT_MAP |
 	 OCFS2_FEATURE_INCOMPAT_INLINE_DATA |
 	 OCFS2_FEATURE_INCOMPAT_META_ECC |
+	 OCFS2_FEATURE_INCOMPAT_XATTR |
+	 OCFS2_FEATURE_INCOMPAT_REFCOUNT_TREE |
+	 OCFS2_FEATURE_INCOMPAT_INDEXED_DIRS |
+	 OCFS2_FEATURE_INCOMPAT_DISCONTIG_BG,
+	 OCFS2_FEATURE_RO_COMPAT_UNWRITTEN |
+	 OCFS2_FEATURE_RO_COMPAT_USRQUOTA |
+	 OCFS2_FEATURE_RO_COMPAT_GRPQUOTA }, /* OCFS2_FEATURE_LEVEL_MAX_FEATURES */
+};
+
+static ocfs2_fs_options mkfstypes_features_defaults[] = {
+	{OCFS2_FEATURE_COMPAT_BACKUP_SB | OCFS2_FEATURE_COMPAT_JBD2_SB,
+	 OCFS2_FEATURE_INCOMPAT_SPARSE_ALLOC |
+	 OCFS2_FEATURE_INCOMPAT_INLINE_DATA |
 	 OCFS2_FEATURE_INCOMPAT_XATTR,
-	 OCFS2_FEATURE_RO_COMPAT_UNWRITTEN}, /* OCFS2_FEATURE_LEVEL_MAX_FEATURES */
+	 OCFS2_FEATURE_RO_COMPAT_UNWRITTEN},  /* OCFS2_MKFSTYPE_DEFAULT */
+
+	{OCFS2_FEATURE_COMPAT_BACKUP_SB | OCFS2_FEATURE_COMPAT_JBD2_SB,
+	 OCFS2_FEATURE_INCOMPAT_SPARSE_ALLOC |
+	 OCFS2_FEATURE_INCOMPAT_INLINE_DATA |
+	 OCFS2_FEATURE_INCOMPAT_XATTR,
+	 OCFS2_FEATURE_RO_COMPAT_UNWRITTEN},  /* OCFS2_MKFSTYPE_DATAFILES */
+
+	{OCFS2_FEATURE_COMPAT_BACKUP_SB | OCFS2_FEATURE_COMPAT_JBD2_SB,
+	 OCFS2_FEATURE_INCOMPAT_SPARSE_ALLOC |
+	 OCFS2_FEATURE_INCOMPAT_INLINE_DATA |
+	 OCFS2_FEATURE_INCOMPAT_XATTR,
+	 OCFS2_FEATURE_RO_COMPAT_UNWRITTEN},  /* OCFS2_MKFSTYPE_MAIL */
+
+	{OCFS2_FEATURE_COMPAT_BACKUP_SB | OCFS2_FEATURE_COMPAT_JBD2_SB,
+	 OCFS2_FEATURE_INCOMPAT_SPARSE_ALLOC |
+	 OCFS2_FEATURE_INCOMPAT_INLINE_DATA |
+	 OCFS2_FEATURE_INCOMPAT_XATTR |
+	 OCFS2_FEATURE_INCOMPAT_REFCOUNT_TREE,
+	 OCFS2_FEATURE_RO_COMPAT_UNWRITTEN},  /* OCFS2_MKFSTYPE_VMSTORE */
 };
 
 /* These are the features we support in mkfs/tunefs via --fs-features */
@@ -122,6 +167,31 @@ static struct fs_feature_flags ocfs2_supported_features[] = {
 		{0, OCFS2_FEATURE_INCOMPAT_XATTR, 0},
 		{0, OCFS2_FEATURE_INCOMPAT_XATTR, 0},
 	},
+	{
+		"indexed-dirs",
+		{0, OCFS2_FEATURE_INCOMPAT_INDEXED_DIRS, 0},
+		{0, OCFS2_FEATURE_INCOMPAT_INDEXED_DIRS, 0},
+	},
+	{
+		"usrquota",
+		{0, 0, OCFS2_FEATURE_RO_COMPAT_USRQUOTA},
+		{0, 0, OCFS2_FEATURE_RO_COMPAT_USRQUOTA},
+	},
+	{
+		"grpquota",
+		{0, 0, OCFS2_FEATURE_RO_COMPAT_GRPQUOTA},
+		{0, 0, OCFS2_FEATURE_RO_COMPAT_GRPQUOTA},
+	},
+	{
+		"refcount",
+		{0, OCFS2_FEATURE_INCOMPAT_REFCOUNT_TREE, 0},
+		{0, OCFS2_FEATURE_INCOMPAT_REFCOUNT_TREE, 0},
+	},
+	{
+		"discontig-bg",
+		{0, OCFS2_FEATURE_INCOMPAT_DISCONTIG_BG, 0},
+		{0, OCFS2_FEATURE_INCOMPAT_DISCONTIG_BG, 0},
+	},
 	{
 		NULL,
 		{0, 0, 0},
@@ -189,6 +259,26 @@ static struct feature_name ocfs2_feature_names[] = {
 		.fn_name = "xattr",
 		.fn_flag = {0, OCFS2_FEATURE_INCOMPAT_XATTR, 0},
 	},
+	{
+		.fn_name = "indexed-dirs",
+		.fn_flag = {0, OCFS2_FEATURE_INCOMPAT_INDEXED_DIRS, 0},
+	},
+	{
+		.fn_name = "usrquota",
+		.fn_flag = {0, 0, OCFS2_FEATURE_RO_COMPAT_USRQUOTA},
+	},
+	{
+		.fn_name = "grpquota",
+		.fn_flag = {0, 0, OCFS2_FEATURE_RO_COMPAT_GRPQUOTA},
+	},
+	{
+		.fn_name = "refcount",
+		.fn_flag = {0, OCFS2_FEATURE_INCOMPAT_REFCOUNT_TREE, 0},
+	},
+	{
+		.fn_name = "discontig-bg",
+		.fn_flag = {0, OCFS2_FEATURE_INCOMPAT_DISCONTIG_BG, 0},
+	},
 	{
 		.fn_name = NULL,
 	},
@@ -214,6 +304,46 @@ static struct tunefs_flag_name ocfs2_tunefs_flag_names[] = {
 	},
 };
 
+/*
+ * The printable names of every flag in e_flags.  If libocfs2 supports the
+ * flag, its name must be here.
+ *
+ * These MUST be kept in sync with the flags in ocfs2_fs.h.
+ */
+static struct extent_flag_name ocfs2_extent_flag_names[] = {
+	{
+		.efn_name = "Unwritten",
+		.efn_flag = OCFS2_EXT_UNWRITTEN,
+	},
+	{
+		.efn_name = "Refcounted",
+		.efn_flag = OCFS2_EXT_REFCOUNTED,
+	},
+	{
+		.efn_name = NULL,
+	},
+};
+
+/*
+ * The printable names of every flag in rf_flags.  If libocfs2 supports the
+ * flag, its name must be here.
+ *
+ * These MUST be kept in sync with the flags in ocfs2_fs.h.
+ */
+static struct refcount_flag_name ocfs2_refcount_flag_names[] = {
+	{
+		.rfn_name = "Leaf",
+		.rfn_flag = OCFS2_REFCOUNT_LEAF_FL,
+	},
+	{
+		.rfn_name = "Tree",
+		.rfn_flag = OCFS2_REFCOUNT_TREE_FL,
+	},
+	{
+		.rfn_name = NULL,
+	},
+};
+
 static inline void merge_features(ocfs2_fs_options *features,
 				  ocfs2_fs_options new_features)
 {
@@ -358,6 +488,90 @@ errcode_t ocfs2_snprint_tunefs_flags(char *str, size_t size, uint16_t flags)
 	return err;
 }
 
+errcode_t ocfs2_snprint_extent_flags(char *str, size_t size, uint8_t flags)
+{
+	int i, printed;
+	char *ptr = str;
+	size_t remain = size;
+	errcode_t err = 0;
+	char *sep = " ";
+	uint8_t found = 0;
+
+	for (i = 0; ocfs2_extent_flag_names[i].efn_name; i++) {
+		if (!(flags & ocfs2_extent_flag_names[i].efn_flag))
+			continue;
+		found |= ocfs2_extent_flag_names[i].efn_flag;
+
+		printed = snprintf(ptr, remain, "%s%s",
+				   ptr == str ? "" : sep,
+				   ocfs2_extent_flag_names[i].efn_name);
+		if (printed < 0)
+			err = OCFS2_ET_INTERNAL_FAILURE;
+		else if (printed >= remain)
+			err = OCFS2_ET_NO_SPACE;
+		if (err)
+			break;
+
+		remain -= printed;
+		ptr += printed;
+	}
+
+	if (!err) {
+		if (found != flags) {
+			printed = snprintf(ptr, remain, "%sUnknown",
+					   ptr == str ? "" : sep);
+			if (printed < 0)
+				err = OCFS2_ET_INTERNAL_FAILURE;
+			else if (printed >= remain)
+				err = OCFS2_ET_NO_SPACE;
+		}
+	}
+
+	return err;
+}
+
+errcode_t ocfs2_snprint_refcount_flags(char *str, size_t size, uint8_t flags)
+{
+	int i, printed;
+	char *ptr = str;
+	size_t remain = size;
+	errcode_t err = 0;
+	char *sep = " ";
+	uint8_t found = 0;
+
+	for (i = 0; ocfs2_refcount_flag_names[i].rfn_name; i++) {
+		if (!(flags & ocfs2_refcount_flag_names[i].rfn_flag))
+			continue;
+		found |= ocfs2_refcount_flag_names[i].rfn_flag;
+
+		printed = snprintf(ptr, remain, "%s%s",
+				   ptr == str ? "" : sep,
+				   ocfs2_refcount_flag_names[i].rfn_name);
+		if (printed < 0)
+			err = OCFS2_ET_INTERNAL_FAILURE;
+		else if (printed >= remain)
+			err = OCFS2_ET_NO_SPACE;
+		if (err)
+			break;
+
+		remain -= printed;
+		ptr += printed;
+	}
+
+	if (!err) {
+		if (found != flags) {
+			printed = snprintf(ptr, remain, "%sUnknown",
+					   ptr == str ? "" : sep);
+			if (printed < 0)
+				err = OCFS2_ET_INTERNAL_FAILURE;
+			else if (printed >= remain)
+				err = OCFS2_ET_NO_SPACE;
+		}
+	}
+
+	return err;
+}
+
 
 /*
  * If we are asked to clear a feature, we also need to clear any other
@@ -385,11 +599,17 @@ static void ocfs2_feature_clear_deps(ocfs2_fs_options *reverse_set)
  * reverse_set: all the features a user want to clear by "--fs-features".
  */
 errcode_t ocfs2_merge_feature_flags_with_level(ocfs2_fs_options *dest,
+					       enum ocfs2_mkfs_types fstype,
 					       int level,
 					       ocfs2_fs_options *feature_set,
 					       ocfs2_fs_options *reverse_set)
 {
-	ocfs2_fs_options level_set = feature_level_defaults[level];
+	ocfs2_fs_options level_set;
+
+	if (level == OCFS2_FEATURE_LEVEL_DEFAULT)
+		level_set = mkfstypes_features_defaults[fstype];
+	else
+		level_set = feature_level_defaults[level];
 
 	/*
 	 * Ensure that all dependancies are correct in the reverse set.
@@ -631,7 +851,42 @@ static void print_tunefs_flags(void)
 	fprintf(stdout, "Printable s_tunefs_flag:\n");
 
 	err = ocfs2_snprint_tunefs_flags(buf, PATH_MAX,
-					 OCFS2_TUNEFS_INPROG_REMOVE_SLOT);
+					 OCFS2_TUNEFS_INPROG_REMOVE_SLOT |
+					 OCFS2_TUNEFS_INPROG_DIR_TRAILER);
+	if (err)
+		snprintf(buf, PATH_MAX, "An error occurred: %s",
+			 error_message(err));
+	fprintf(stdout, "FLAGS:\t\t%s\n", buf);
+	fprintf(stdout, "\n");
+}
+
+static void print_extent_flags(void)
+{
+	errcode_t err;
+	char buf[PATH_MAX];
+
+	fprintf(stdout, "Printable e_flags:\n");
+
+	err = ocfs2_snprint_extent_flags(buf, PATH_MAX,
+					 OCFS2_EXT_UNWRITTEN |
+					 OCFS2_EXT_REFCOUNTED);
+	if (err)
+		snprintf(buf, PATH_MAX, "An error occurred: %s",
+			 error_message(err));
+	fprintf(stdout, "FLAGS:\t\t%s\n", buf);
+	fprintf(stdout, "\n");
+}
+
+static void print_refcount_flags(void)
+{
+	errcode_t err;
+	char buf[PATH_MAX];
+
+	fprintf(stdout, "Printable rf_flags:\n");
+
+	err = ocfs2_snprint_refcount_flags(buf, PATH_MAX,
+					   OCFS2_REFCOUNT_TREE_FL |
+					   OCFS2_REFCOUNT_LEAF_FL);
 	if (err)
 		snprintf(buf, PATH_MAX, "An error occurred: %s",
 			 error_message(err));
@@ -717,7 +972,9 @@ int main(int argc, char *argv[])
 	}
 
 	memset(&mkfs_features, 0, sizeof(ocfs2_fs_options));
-	err = ocfs2_merge_feature_flags_with_level(&mkfs_features, level,
+	err = ocfs2_merge_feature_flags_with_level(&mkfs_features,
+						   OCFS2_MKFSTYPE_DEFAULT,
+						   level,
 						   &set_features,
 						   &clear_features);
 	if (err) {
@@ -738,6 +995,8 @@ int main(int argc, char *argv[])
 	print_order(1, &clear_features);
 
 	print_tunefs_flags();
+	print_extent_flags();
+	print_refcount_flags();
 
 	return 0;
 }
