@@ -31,6 +31,7 @@
 #include <string.h>
 #include <inttypes.h>
 #include <errno.h>
+#include <stdbool.h>
 
 /* I hate glibc and gcc */
 #ifndef ULLONG_MAX
@@ -48,10 +49,11 @@
  * function. At this point this function returns EIO if image file has any
  * holes
  */
-inline errcode_t ocfs2_read_blocks(ocfs2_filesys *fs, int64_t blkno,
-		int count, char *data)
+static errcode_t __ocfs2_read_blocks(ocfs2_filesys *fs, int64_t blkno,
+				     int count, char *data, bool nocache)
 {
 	int i;
+	errcode_t err;
 
 	if (fs->fs_flags & OCFS2_FLAG_IMAGE_FILE) {
 		/*
@@ -62,11 +64,29 @@ inline errcode_t ocfs2_read_blocks(ocfs2_filesys *fs, int64_t blkno,
 		 */
 		for (i = 0; i < count; i++)
 			if (!ocfs2_image_test_bit(fs, blkno+i))
-				return -EIO;
+				return OCFS2_ET_IO;
 		/* translate the block number */
 		blkno = ocfs2_image_get_blockno(fs, blkno);
 	}
-	return io_read_block(fs->fs_io, blkno, count, data);
+
+	if (nocache)
+		err = io_read_block_nocache(fs->fs_io, blkno, count, data);
+	else
+		err = io_read_block(fs->fs_io, blkno, count, data);
+
+	return err;
+}
+
+errcode_t ocfs2_read_blocks_nocache(ocfs2_filesys *fs, int64_t blkno,
+				    int count, char *data)
+{
+	return __ocfs2_read_blocks(fs, blkno, count, data, true);
+}
+
+errcode_t ocfs2_read_blocks(ocfs2_filesys *fs, int64_t blkno,
+			    int count, char *data)
+{
+	return __ocfs2_read_blocks(fs, blkno, count, data, false);
 }
 
 static errcode_t ocfs2_validate_ocfs1_header(ocfs2_filesys *fs)
@@ -137,7 +157,7 @@ errcode_t ocfs2_read_super(ocfs2_filesys *fs, uint64_t superblock, char *sb)
 	orig_blocksize = fs->fs_blocksize;
 	fs->fs_super = (struct ocfs2_dinode *)swapblk;
 	fs->fs_blocksize = blocksize;
-	ocfs2_swap_inode_to_cpu(fs->fs_super, fs->fs_blocksize);
+	ocfs2_swap_inode_to_cpu(fs, fs->fs_super);
 
 	ret = ocfs2_validate_meta_ecc(fs, blk, &di->i_check);
 
@@ -148,7 +168,7 @@ errcode_t ocfs2_read_super(ocfs2_filesys *fs, uint64_t superblock, char *sb)
 	if (ret)
 		goto out_blk;
 
-	ocfs2_swap_inode_to_cpu(di, fs->fs_blocksize);
+	ocfs2_swap_inode_to_cpu(fs, di);
 	if (!sb)
 		fs->fs_super = di;
 	else {
