@@ -27,7 +27,6 @@
 #include <asm/types.h>
 #include <inttypes.h>
 
-#include <asm/page.h>
 #include <sys/mount.h>
 #include <dirent.h>
 #include <stdio.h>
@@ -49,10 +48,6 @@
 #include "mount_constants.h"
 #include "opts.h"
 
-#include <ocfs2.h>
-#include <ocfs2_fs.h>
-#include <ocfs1_fs_compat.h>
-#include <kernel-list.h>
-
-#include "bitops.h"
+#include "ocfs2/ocfs2.h"
+#include "ocfs2/bitops.h"
 
