@@ -38,7 +38,7 @@
 #include "jiterator.h"
 #include "o2cb_config.h"
 
-#include "o2cb.h"
+#include "o2cb/o2cb.h"
 
 #define PROGNAME "o2cb_ctl"
 #define O2CB_CONFIG_FILE "/etc/ocfs2/cluster.conf"
@@ -904,7 +904,7 @@ static gint online_cluster(O2CBContext *ctxt, O2CBCluster *cluster)
         ret = O2CB_ET_CONFIGURATION_ERROR;
         com_err(PROGNAME, ret, "while populating cluster %s.  None of its "
                 "nodes were considered local.  A node is considered local "
-                "when its node name in the configuration maches this "
+                "when its node name in the configuration matches this "
                 "machine's host name.", name);
         rc = -EINVAL;
         goto out_error;
@@ -1265,7 +1265,7 @@ static gint run_create_nodes(O2CBContext *ctxt)
         }
 
         tmpnode = o2cb_cluster_get_node(cluster, num);
-        if (tmpnode)
+        if (tmpnode && (tmpnode != node))
         {
             rc = -EEXIST;
             fprintf(stderr,
@@ -1377,9 +1377,9 @@ gint main(gint argc, gchar *argv[])
 
     ret = o2cb_init();
     if (ret) {
-	com_err(PROGNAME, ret, "Cannot initialize cluster\n");
-	rc = -EINVAL;
-	goto out_error;
+	    com_err(PROGNAME, ret, "Cannot initialize cluster\n");
+	    rc = -EINVAL;
+	    goto out_error;
     }
 
     switch (ctxt.oc_op)
