@@ -35,6 +35,7 @@
 #include <signal.h>
 #include <syslog.h>
 #include <sched.h>
+#include <limits.h>
 
 #include "ocfs2-kernel/kernel-list.h"
 #include "o2cb/o2cb.h"
@@ -204,7 +205,7 @@ static int setup_sigpipe(void)
 {
 	int rc;
 	int signal_pipe[2];
-	struct sigaction act;
+	struct sigaction act = { .sa_handler = handler, };
 
 	rc = pipe(signal_pipe);
 	if (rc) {
@@ -216,10 +217,7 @@ static int setup_sigpipe(void)
 
 	sigpipe_write_fd = signal_pipe[1];
 
-	act.sa_sigaction = NULL;
-	act.sa_restorer = NULL;
 	sigemptyset(&act.sa_mask);
-	act.sa_handler = handler;
 #ifdef SA_INTERRUPT
 	act.sa_flags = SA_INTERRUPT;
 #endif
@@ -245,44 +243,6 @@ out:
 	return rc;
 }
 
-int do_read(int fd, void *buf, size_t count)
-{
-	int rv, off = 0;
-
-	while (off < count) {
-		rv = read(fd, buf + off, count - off);
-		if (rv == 0)
-			return -1;
-		if (rv == -1 && errno == EINTR)
-			continue;
-		if (rv == -1)
-			return -1;
-		off += rv;
-	}
-	return 0;
-}
-
-int do_write(int fd, void *buf, size_t count)
-{
-	int rv, off = 0;
-
- retry:
-	rv = write(fd, buf + off, count);
-	if (rv == -1 && errno == EINTR)
-		goto retry;
-	if (rv < 0) {
-		log_error("write errno %d", errno);
-		return rv;
-	}
-
-	if (rv != count) {
-		count -= rv;
-		off += rv;
-		goto retry;
-	}
-	return 0;
-}
-
 static int do_mount(int ci, int fd, const char *fstype, const char *uuid,
 		    const char *cluster, const char *device,
 		    const char *service)
@@ -406,18 +366,84 @@ int connection_add(int fd, void (*work)(int ci), void (*dead)(int ci))
 	return -ELOOP;
 }
 
-static int dump_debug(int ci)
+/* 4 characters for "ITEM", 1 for the space, 1 for the '\0' */
+#define DEBUG_BYTES_PER_ITEM	(OCFS2_CONTROLD_MAXLINE - 6)
+static size_t debug_bytes_to_count(size_t bytes)
 {
-	int len = DUMP_SIZE;
+	return (bytes + DEBUG_BYTES_PER_ITEM - 1) / DEBUG_BYTES_PER_ITEM;
+}
 
-	if (dump_wrap) {
-		len = DUMP_SIZE - dump_point;
-		do_write(client[ci].fd, dump_buf + dump_point, len);
-		len = dump_point;
+static int send_debug(int fd, const char *ptr, size_t bytes)
+{
+	int rc = 0;
+	size_t remain = bytes;
+	size_t itemlen = DEBUG_BYTES_PER_ITEM;
+	char itembuf[DEBUG_BYTES_PER_ITEM + 1];
+
+	while (remain) {
+		if (itemlen > remain)
+			itemlen = remain;
+		memcpy(itembuf, ptr, itemlen);
+		itembuf[itemlen] = '\0';
+		rc = send_message(fd, CM_ITEM, itembuf);
+		if (rc)
+			break;
+
+		ptr += itemlen;
+		remain -= itemlen;
 	}
 
-	do_write(client[ci].fd, dump_buf, len);
-	return 0;
+	return rc;
+}
+
+static int dump_debug(int ci, int fd)
+{
+	int rc, rctmp;
+	char error_msg[100];  /* Arbitrary size smaller than a message */
+	int count = 0;
+
+	if (dump_wrap)
+		count += debug_bytes_to_count(DUMP_SIZE - dump_point);
+	count += debug_bytes_to_count(dump_point);
+
+	rc = send_message(fd, CM_ITEMCOUNT, count);
+	if (rc) {
+		snprintf(error_msg, sizeof(error_msg),
+			 "Unable to send ITEMCOUNT: %s",
+			 strerror(-rc));
+		goto out_status;
+	}
+
+	if (dump_wrap) {
+		rc = send_debug(fd, dump_buf + dump_point,
+				DUMP_SIZE - dump_point);
+		if (rc) {
+			snprintf(error_msg, sizeof(error_msg),
+				 "Unable to send ITEM: %s",
+				 strerror(-rc));
+			goto out_status;
+		}
+	}
+	rc = send_debug(fd, dump_buf, dump_point);
+	if (rc) {
+		snprintf(error_msg, sizeof(error_msg),
+			 "Unable to send ITEM: %s",
+			 strerror(-rc));
+		goto out_status;
+	}
+
+	strcpy(error_msg, "OK");
+
+out_status:
+	rctmp = send_message(fd, CM_STATUS, -rc, error_msg);
+	if (rctmp) {
+		log_error("Error sending STATUS message: %s",
+			  strerror(-rc));
+		if (!rc)
+			rc = rctmp;
+	}
+
+	return rc;
 }
 
 static int send_filesystems(int ci, int fd, const char *fstype,
@@ -540,6 +566,10 @@ static void process_client(int ci)
 		rv = send_filesystems(ci, fd, argv[0], argv[1]);
 		break;
 
+		case CM_DUMP:
+		rv = dump_debug(ci, fd);
+		break;
+
 		case CM_STATUS:
 		log_error("Someone sent us cm_status!");
 		break;
@@ -548,20 +578,6 @@ static void process_client(int ci)
 		log_error("Invalid message received");
 		break;
 	}
-#if 0
-	if (daemon_debug_opt)
-		dump_state();
-#endif
-
-#if 0
-	} else if (!strcmp(cmd, "dump")) {
-		dump_debug(ci);
-
-	} else {
-		rv = -EINVAL;
-		goto reply;
-	}
-#endif
 
 	return;
 }
@@ -575,7 +591,7 @@ static void process_listener(int ci)
 		return;
 	}
 
-	i = connection_add(fd, process_client, NULL);
+	i = connection_add(fd, process_client, dead_client);
 	if (i < 0) {
 		log_error("Error adding client: %s", strerror(-i));
 		close(fd);
@@ -797,13 +813,16 @@ static int protocol_compatible(struct ocfs2_protocol_version *cluster,
 
 static int read_global_checkpoint(void)
 {
-	int rc;
+	int rc, seen = 0, opened = 0, retrycount = 0;
 	char *buf;
 	size_t len;
 
+retry:
 	rc = ckpt_open_global(0);
 	if (rc)
 		goto out;
+	seen = 1;
+	opened = 1;
 
 	rc = ckpt_global_get(DAEMON_PROTOCOL_SECTION, &buf, &len);
 	if (rc)
@@ -846,6 +865,38 @@ static int read_global_checkpoint(void)
 	}
 
 out:
+	if (rc == -ENOENT) {
+		/*
+		 * -ENOENT means the first daemon hasn't gotten the
+		 * global checkpoint fully installed yet.  Either the
+		 * checkpoint or one of its sections is missing.
+		 *
+		 * If we saw the checkpoint once, but now it's gone, it
+		 * means the first daemon died.  We don't recover from
+		 * that.  But if we haven't seen the checkpoint yet, or if
+		 * it's just a missing section, we can keep trying.
+		 */
+
+		if (opened)
+			ckpt_close_global();
+
+		if (seen && !opened) {
+			log_error("The global checkpoint disappeared out "
+				  "from underneath us.  This shouldn't "
+				  "happen to a daemon that is not the "
+				  "first in the cluster");
+		} else {
+			opened = 0;
+			retry_warning(retrycount,
+				      "Attempted to read the cluster's "
+				      "protocol versions %d times, still "
+				      "trying",
+				      retrycount);
+			sleep_ms(10);
+			goto retry;
+		}
+	}
+
 	return rc;
 }
 
@@ -873,6 +924,13 @@ static void cpg_joined(int first)
 	log_debug("fs protocol is %d.%d",
 		  fs_running_proto.pv_major, fs_running_proto.pv_minor);
 
+	log_debug("Connecting to dlm_controld");
+	rv = setup_dlmcontrol();
+	if (rv) {
+		shutdown_daemon();
+		return;
+	}
+
 	log_debug("Opening control device");
 	err = o2cb_control_open(our_nodeid, &fs_running_proto);
 	if (err) {
@@ -898,7 +956,7 @@ static int loop(void)
 	if (rv < 0)
 		goto out;
 
-	rv = setup_cman();
+	rv = setup_stack();
 	if (rv < 0)
 		goto out;
 
@@ -951,10 +1009,11 @@ stop:
 	bail_on_mounts();
 
 	o2cb_control_close();
+	exit_dlmcontrol();
 	exit_cpg();
 	drop_node_checkpoint();
 	exit_ckpt();
-	exit_cman();
+	exit_stack();
 
 out:
 	return rv;
@@ -1004,6 +1063,7 @@ static void lockfile(void)
 
 static void daemonize(void)
 {
+	int fd;
 	pid_t pid = fork();
 	if (pid < 0) {
 		perror("main: cannot fork");
@@ -1017,6 +1077,18 @@ static void daemonize(void)
 	close(0);
 	close(1);
 	close(2);
+	fd = open("/dev/null", O_RDWR);
+	if (fd >= 0) {
+		/* dup2 to 0 / 1 / 2 (stdin / stdout / stderr) */
+		dup2(fd, STDIN_FILENO);  /* 0 */
+		dup2(fd, STDOUT_FILENO); /* 1 */
+		dup2(fd, STDERR_FILENO); /* 2 */
+
+		/* Should be 0, but just in case it isn't... */
+		if (fd > 2) {
+			close(fd);
+		}
+	}
 	openlog("ocfs2_controld", LOG_PID, LOG_DAEMON);
 
 	lockfile();
@@ -1128,9 +1200,9 @@ int main(int argc, char **argv)
 		com_err(prog_name, err, "while determining the current cluster stack");
 		return 1;
 	}
-	if (strcmp(stack, "cman")) {
-		fprintf(stderr, "%s: This daemon supports the \"cman\" stack, but the \"%s\" stack is in use\n",
-			prog_name, stack);
+	if (strcmp(stack, stackname)) {
+		fprintf(stderr, "%s: This daemon supports the \"%s\" stack, but the \"%s\" stack is in use\n",
+			prog_name, stackname, stack);
 		return 1;
 	}
 
